#include <stdio.h>
#include <libspe.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>

#define SPE_FILE "./ex4core_speprev"

#define SPE_STOP  0
#define SPE_START 1
#define SPE_DONE  2

#define PPEMM_DONE  0
#define PPEMM_DOING 1

#define SPE_NUM  6

/* default m, n */
#define DM 1000
#define DN 1000

typedef struct {
  volatile double *result_p;
  volatile int flag;
  long int random_seed;
  int size;
} __attribute__ ((aligned(128))) dma_data_t;
static volatile dma_data_t dma_data[SPE_NUM];

typedef struct {
  volatile int flag;
  volatile int size;
} buf_data_t;
static volatile buf_data_t buf_data[SPE_NUM];

static volatile double result[2][SPE_NUM][2048] __attribute__ ((aligned(128)));

int create_matrix(double** mat, int* m, int* n);

void get_dma_size(int* remain, int* dma_size)
{
  *dma_size = 256;
  if(*remain < 256){ *dma_size = 128;}
  if(*remain < 128){ *dma_size = 64;}
  if(*remain < 64){ *dma_size = 16;}
  if(*remain < 16){ *dma_size = 8; }
  if(*remain < 8) { *dma_size = *remain; }
  //  if(*remain < 4) { *dma_size = 2; }
  //  if(*remain < 2) { *dma_size = 1; }
  *remain -= *dma_size;
}

void get_dma_size_gp(int *remain, int* dma_size, int sp)
{
  *dma_size = 2024;
  if(*remain < 2048*sp){ *dma_size = 1024;}
  if(*remain < 1024*sp){ *dma_size = 512;}
  if(*remain < 512*sp){ *dma_size = 256;}
  if(*remain < 256*sp){ *dma_size = 128;}
  if(*remain < 128*sp){ *dma_size = -1; return;}
  //if(*remain < 64*sp){ *dma_size = 16;}
  //if(*remain < 16*sp){ *dma_size = 8; }
  //if(*remain < 8*sp) { *dma_size = -1; return;}
  *remain -= ((*dma_size)*sp);
}  

#ifdef LIBMAIN
int main(int argc, void **argv)
{
  double *mat;
  int m, n;

  int lc, lc2;

  
  if(create_matrix(&mat, &m, &n) == -1){
    printf("No memory...\n");
    exit(0);
  }
  
  for(lc = 0; lc < m; lc++){
    for(lc2 = 0; lc2 < n; lc2++){
      printf("%.3lf ", mat[lc*m+lc2]);
    }
    printf("\n");
  }
  
}
#endif

int create_matrix(double** mat, int* m, int* n)
{
  int i;
  spe_program_handle_t *spe_handle[SPE_NUM];
  speid_t spe_id[SPE_NUM];
  int status[SPE_NUM];
  int size[SPE_NUM];
  int lm, ln;
  int gen, remain, mbuf, tbuf;
  int jobsize;
  double *res_p;
  
  /* --- PPE process --- */
  *m = DM;
  *n = DN;
  lm = DM;
  ln = DN;
  remain = lm * ln;
  *mat = (double*) calloc(lm*ln, sizeof(double));
  res_p = *mat;
  if(*mat == NULL){ return -1;}

  srand(time(NULL));
    
  /* --- SPE process --- */
  //return -1;



  while(remain > 0){
    tbuf = gen % 2;
    mbuf = (gen+1) % 2;
    //printf("buffer manage mbuf=%p, tbuf=%p\n", &(result[tbuf][i][0]), &(result[mbuf][i][0]));


    //sync size div//
    get_dma_size_gp(&remain, &size[0], SPE_NUM);
	

    for (i = 0, jobsize = 0; (i < SPE_NUM) && (remain > 0); i++) {
      //async size div//
      //get_dma_size(&remain, &size[i]);
      if(size[0] > 0){
	size[i] = size[0]; 
      }else{
	size[0] = remain;
	remain = 0;
      }
      dma_data[i].flag = SPE_STOP;
      dma_data[i].size = size[i];
      dma_data[i].result_p = &(result[mbuf][i][0]);
      dma_data[i].random_seed = (long) rand();
      //printf("remain=%d\n", remain);
      jobsize++;
    }
    
    //printf("Open spe program\n");
    for (i = 0; i < jobsize; i++) {
      spe_handle[i] = spe_open_image(SPE_FILE);
      if (spe_handle[i] == 0) {
      printf("[ERR] Can't open spe%d program\n", i);
      return -1;
      }
    }    
    //printf("Create and Start spe program\n");

    for (i = 0; i < jobsize; i++) {
      spe_id[i] =
	spe_create_thread(0, spe_handle[i], (void *) &dma_data[i],
			  NULL, -1, 0);
      if (spe_id[i] == 0) {
	printf("[ERR] Can't create spe%d thread\n", i);
	return -1;
      }
    }
    
    for (i = 0; i < jobsize; i++) {
      dma_data[i].flag = SPE_START;
      //printf("%d:flag=%d\n", i, dma_data[i].flag);
    }
    
    for (i = 0; (i < SPE_NUM) && (buf_data[i].flag == PPEMM_DOING); i++){
      memcpy(res_p, &(result[tbuf][i][0]), sizeof(double)*buf_data[i].size);
      res_p += buf_data[i].size;
      buf_data[i].flag = PPEMM_DONE;
      //printf("write from %p to %p size %d\n", &(result[tbuf][i][0]), res_p, buf_data[i].size); 
    }
    
    //printf("waiting spe...\n");
 
    for (i = 0; i < jobsize; i++) {
      while (dma_data[i].flag != SPE_DONE);
 
      buf_data[i].flag = PPEMM_DOING;
      buf_data[i].size = dma_data[i].size;
      
    }
    
    for (i = 0; i < jobsize; i++) {
      //printf("Waiting spe%d program is finish...", i);
      spe_wait(spe_id[i], &status[i], 0);
      //printf("done\n");
      
      //printf("Relese spe%d program\n", i);
      spe_close_image(spe_handle[i]);
    }

    if(remain <= 0){
      //printf("finalize\n");
      for (i = 0; i < jobsize; i++){
	memcpy(res_p, &(result[mbuf][i][0]), sizeof(double)*buf_data[i].size);
	res_p += buf_data[i].size;
	buf_data[i].flag = PPEMM_DONE;
	//printf("write from %p to %p size %d\n", &(result[mbuf][i][0]), res_p, buf_data[i].size); 
      }
      
    }
    gen++;
  }
  
  printf("START LU decomp / Time=%ld.\n", time(NULL));
  
  return 0;
}
